import asyncio
import asyncssh

import sys
from threading import Thread

import time
import json
import logging
import gzip


class _SSHClient(asyncssh.SSHClient):
    def connection_made(self, conn):
        logging.debug(f"Connection made to conn.get_extra_info('peername')[0]")

    def auth_completed(self):
        logging.debug('Authentication successful')


class _SSHClientSession(asyncssh.SSHTCPSession):

    def connection_made(self, chan):
        logging.debug("Session opened")
        self._chan = chan
        self._requests = dict()

    def connection_lost(self, exc):
        logging.debug("Connection lost")
        logging.debug(f"{exc}")

    def session_started(self):
        logging.debug("Session successful")

    def data_received(self, data, datatype):
        logging.debug(f"Received data: {data}")

        try:
            data = json.loads(gzip.decompress(data).decode('utf-8'))

            if data['request_id'] in self._requests:
                if callable(self._requests[data['request_id']]):
                    self._requests[data['request_id']](data)

                if self._requests[data['request_id']] is None:
                    self._requests[data['request_id']] = data
                else:
                    del(self._requests[data['request_id']])

        except Exception:
            logging.exception(f"There was an error processing the server response")

    def eof_received(self):
        logging.debug("Received EOF")
        self._chan.exit(0)

    async def send_request(self, verb, resource, body, callback):
        if verb not in ['GET', 'STORE', 'UPDATE', 'DELETE']:
            raise ValueError("Unknown verb")

        request = {
            'id': time.time(),
            'verb': verb,
            'resource': resource,
            'body': body
        }

        self._requests[request['id']] = callback
        self._chan.write(gzip.compress(json.dumps(request, separators=[',', ':']).encode('utf-8')))
        logging.debug(f"{verb} {resource} {body}")

        return request['id']


class KorvClient:

    def __init__(self, host='localhost', port=8022, client_keys=None, known_hosts=None, max_packet_size=32768):
        self.max_packet_size = max_packet_size

        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._session = asyncio.get_event_loop().run_until_complete(self.__connect(host, port, known_hosts, client_keys))

        try:
            t = Thread(target=self.__start_loop, args=(self._loop,))
            t.start()
            # asyncio.run_coroutine_threadsafe(self.__connect(), self._loop)

        except (OSError, asyncssh.Error) as exc:
            sys.exit(f'SSH connection failed: {exc}')

    def __start_loop(self, loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    async def __connect(self, host, port, known_hosts, client_keys):
        logging.info(f"Connecting to SSH Server {host}:{port}")
        conn, client = await asyncssh.create_connection(
            _SSHClient,
            host,
            port,
            client_keys=client_keys,
            known_hosts=known_hosts
        )

        logging.debug("Opening Socket")
        chan, session = await conn.create_connection(_SSHClientSession, host, port, max_pktsize=self.max_packet_size)
        return session

    def get(self, resource, body=None, callback=None):
        if callback is None:
            request_id = asyncio.run_coroutine_threadsafe(self._session.send_request("GET", resource, body, None), self._loop).result()

            while self._session._requests[request_id] is None:
                time.sleep(0.1)

            response = self._session._requests[request_id]
            del(self._session._requests[request_id])

            return response
        else:
            asyncio.run_coroutine_threadsafe(self._session.send_request("GET", resource, body, callback), self._loop)

    def store(self, resource, body, callback=None):
        asyncio.run_coroutine_threadsafe(self._session.send_request("STORE", resource, body, callback), self._loop)

    def update(self, resource, body, callback=None):
        asyncio.run_coroutine_threadsafe(self._session.send_request("UPDATE", resource, body, callback), self._loop)

    def delete(self, resource, body=None, callback=None):
        asyncio.run_coroutine_threadsafe(self._session.send_request("DELETE", resource, body, callback), self._loop)



import asyncio
import asyncssh
import logging

import sys
import traceback
import time
import json
import gzip


class _KorvServerSession(asyncssh.SSHTCPSession):
    def __init__(self, callbacks):
        self._callbacks = callbacks

    def connection_made(self, chan):
        """New connection established"""

        logging.info("Connection incoming")
        self._chan = chan

    def connection_lost(self, exc):
        """Lost the connection to the client"""

        logging.info("Connection lost")
        logging.info(f"{exc}")

    def session_started(self):
        """New session established succesfully"""

        logging.info("Connection successful")

    def data_received(self, data, datatype):
        """New data coming in"""

        logging.info(f"Received data: {data}")
        self._dispatch(data)

    def eof_received(self):
        """Got an EOF, close the channel"""

        logging.info("EOF")
        self._chan.exit(0)

    def _dispatch(self, data):
        try:
            request = json.loads(gzip.decompress(data).decode('utf-8'))

            if 'id' not in request:
                logging.info("Malformed request: missing 'id'")
                self._send_response(0, 400, {"message": "Missing 'id'"})

            if 'verb' not in request:
                logging.info("Malformed request: missing 'request'")
                self._send_response(request['id'], 400, {"message": "Missing 'verb'"})

            if 'resource' not in request:
                logging.info("Malformed request: missing 'resource'")
                self._send_response(request['id'], 400, {"message": "Missing 'resource'"})

            if request['verb'] == 'STORE' and 'body' not in request['request']:
                logging.info("Malformed request: missing 'resource'")
                self._send_response(request['id'], 400, {"message": "Missing 'body'"})

            elif request['verb'] == 'UPDATE' and 'body' not in request['request']:
                logging.info("Malformed request: missing 'resource'")
                self._send_response(request['id'], 400, {"message": "Missing 'body'"})

        except Exception as e:
            logging.info("Unable to process request")
            self._send_response(0, 400, {"message": "Unable to process request"})

        self.__process_request(request)

    def __process_request(self, request):
        if request['verb'] not in self._callbacks:
            logging.info(f"No callback found for {request['verb']}")
            self._send_response(request['id'], 404)
            return

        if request['resource'] not in self._callbacks[request['verb']]:
            logging.info(f"No callback found for {request['verb']} on {request['resource']}")
            self._send_response(request['id'], 404)
            return

        for callback in self._callbacks[request['verb']][request['resource']]:
            try:
                self._send_response(request['id'], *callback(request))

            except Exception as e:
                logging.info(f"Internal error when executing {request['verb']} on {request['resource']}")
                self._send_response(request['id'], 500, {"message": str(e), "traceback": traceback.format_exc()})

    def _send_response(self, request_id, code, body=None):
        """Send a response to the given client request"""

        cmd = {
            'id': time.time(),
            'request_id': request_id,
            'code': code,
            'body': body
        }

        logging.info(f"Sending response {cmd}")
        self._chan.write(gzip.compress(json.dumps(cmd, separators=[',', ':']).encode('utf-8')))


class KorvServer(asyncssh.SSHServer):
    VERBS = ('GET', 'STORE', 'UPDATE', 'DELETE')

    _callbacks = {verb: dict() for verb in VERBS}

    def __init__(self, port=8022, host_keys=['ssh_host_key'], authorized_client_keys='authorized_keys'):
        """Instatiate an SSH server that listens on the given port for clients that match the authorized keys"""

        self.port = port
        self._host_keys = host_keys
        self._authorized_client_keys = authorized_client_keys

    def connection_requested(self, dest_host, dest_port, orig_host, orig_port):
        """Run a new TCP session that handles an SSH client connection"""

        logging.info(f"Connection requested {dest_host} {dest_port} {orig_host} {orig_port}")
        return _KorvServerSession(KorvServer._callbacks)

    async def __create_server(self):
        """Creates an asynchronous SSH server"""

        await asyncssh.create_server(
            KorvServer, '', self.port,
            server_host_keys=self._host_keys,
            authorized_client_keys=self._authorized_client_keys
        )

    def add_callback(self, verb, resource, callback):
        """Configure a callable to execute when receiving a request with the given verb and resource combination"""

        if verb not in KorvServer.VERBS:
            raise ValueError(f"Verb must be one of {KorvServer.VERBS}")

        if resource not in KorvServer._callbacks[verb]:
            KorvServer._callbacks[verb][resource] = list()

        KorvServer._callbacks[verb][resource].append(callback)

    def start(self):
        """Start the server"""

        logging.info(f"Listening on port {self.port}")

        loop = asyncio.get_event_loop()

        try:
            loop.run_until_complete(self.__create_server())

        except (OSError, asyncssh.Error) as exc:
            sys.exit(f'Error starting server: {exc}')

        loop.run_forever()



from setuptools import setup, find_packages
from os import path


# Get the long description from the README file
with open(path.join('.', 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    author = "tryexceptpass",
    author_email = "cmedina@tryexceptpass.org",

    name = "korv",
    version = "0.2.1",
    description = "SSH API Framework",
    long_description=long_description,
    long_description_content_type='text/markdown',

    url = "https://github.com/tryexceptpass/korv",

    packages = find_packages(),

    install_requires = ['asyncssh'],
    python_requires='>=3.6',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

    license = "MIT",
    classifiers = [ 'License :: OSI Approved :: MIT License',

                    'Topic :: Communications',
                    'Topic :: Internet',

                    'Framework :: AsyncIO',

                    'Programming Language :: Python :: 3 :: Only',
                    'Programming Language :: Python :: 3.6',
                    'Programming Language :: Python :: 3.7',

                    'Development Status :: 4 - Beta',
                  ],

    keywords = 'ssh api framework',

    project_urls={
        'Gitter Chat': 'https://gitter.im/try-except-pass/korv',
        'Say Thanks!': 'https://saythanks.io/to/tryexceptpass',
        'Source': 'https://github.com/tryexceptpass/korv',
        # 'Documentation': 'http://korv.readthedocs.io/en/latest/',
    },
)
